<?php

declare(strict_types=1);

namespace Drupal\generic_layout;

use Drupal;
use Sabberworm\CSS\OutputFormat;

use Drupal\Core\File\FileSystemInterface;
use Drupal\generic_layout\GenericLayoutHelper;

use Drupal\Core\Asset\LibraryDiscoveryCollector;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\layout_builder\LayoutEntityHelperTrait;
use Drupal\Core\Plugin\Context\EntityContext;

use Drupal\layout_builder\LayoutTempstoreRepository;
use Drupal\generic_layout\Plugin\Layout\GenericLayout;
use Drupal\Core\Entity\EntityDisplayRepositoryInterface;
use Drupal\layout_builder\Entity\LayoutBuilderEntityViewDisplay;


class GenericLayoutManager {
  use LayoutEntityHelperTrait;
  const DIRECTORY = "public://generic_layout";

  public function __construct(
    protected readonly EntityTypeManagerInterface $entityTypeManager,
    protected readonly EntityTypeBundleInfoInterface $entityTypeBundleInfo,
    protected readonly EntityDisplayRepositoryInterface $entityDisplayRepository,
    protected readonly FileSystemInterface $fileSystem,
    protected readonly LibraryDiscoveryCollector $libraryDiscoveryCollector,
    protected readonly LayoutTempstoreRepository $layoutTempstoreRepository,
  ) {
  }

  public static function getCssUri(GenericLayout $layout): string {
    $subfolder = $layout->getPluginDefinition()->id();
    $uri = "public://{$subfolder}/{$layout->getInstance()}.css";

    return $uri;
  }

  public function getGenericLayoutSections(): array {
    $section_storage = [];

    foreach ($this->entityTypeManager->getDefinitions() as $entity_type_id => $info) {
      $entityStorage = $this->entityTypeManager->getStorage($entity_type_id);
      foreach ($this->entityTypeBundleInfo->getBundleInfo($entity_type_id) as $bundle => $bundle_info) {
        // Collect Defaults
        foreach(array_keys($this->entityDisplayRepository->getViewModeOptionsByBundle($entity_type_id, $bundle)) as $view_mode) {
          $display = $this->entityDisplayRepository->getViewDisplay($entity_type_id, $bundle, $view_mode);
          if($display instanceof LayoutBuilderEntityViewDisplay && $display->isLayoutBuilderEnabled()) {
            $test = $this->sectionStorageManager()->load('defaults', ['display' => EntityContext::fromEntity($display)]);
            $section_storage[] = $this->sectionStorageManager()->load('defaults', ['display' => EntityContext::fromEntity($display)]);
            if($display->isOverridable()) {
              $entity_ids = $entityStorage->getQuery()->accessCheck(FALSE)->execute();
              foreach ($entityStorage->loadMultiple($entity_ids) as $entity) {
                // @return Null: entity uses default Layout / is not overriden
                if($ss = $this->getSectionStorageForEntity($entity, $view_mode)) {
                  $section_storage[] = $ss;
                }
              }
            }
          }
        }
      }
    }

    $sections = [];
    foreach($section_storage as $ss) {
      $sections = array_merge($sections, GenericLayoutHelper::extractGenericLayoutSections($ss->getSections()));

      if ($this->layoutTempstoreRepository->has($ss)) {
        $sst = $this->layoutTempstoreRepository->get($ss);
        $sections = array_merge($sections, GenericLayoutHelper::extractGenericLayoutSections($sst->getSections()));
      }
    }

    return $sections;
  }

  // @see: https://github.com/sabberworm/PHP-CSS-Parser#use-parser-to-prepend-an-id-to-all-selectors
  public static function prefixSelectors(string $prefix, \Sabberworm\CSS\CSSList\Document &$css) {
    foreach ($css->getAllDeclarationBlocks() as $block) {
      foreach ($block->getSelectors() as $selector) {
        $selector->setSelector($prefix . ' ' . $selector->getSelector());
      }
    }
  }

  public function storeCssFile(GenericLayout $layout): string {
    $instance = $layout->getInstance();
    $css_uri = self::getCssUri($layout);

    $css = $layout->getCss();
    self::prefixSelectors('.' . $instance, $css);
    $rendered_css = $css->render(OutputFormat::createPretty());

    $directory = self::DIRECTORY;
    $this->fileSystem->prepareDirectory($directory, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS);

    return $this->fileSystem->saveData($rendered_css, $css_uri, FileSystemInterface::EXISTS_REPLACE);
  }

  public function removeCssFile(GenericLayout $layout): void {
    $css_uri = self::getCssUri($layout);
    $this->fileSystem->delete($css_uri);
  }

  public function removeAbandonedCssFiles(array $files_to_keep): void {
    $files = $this->fileSystem->scanDirectory(self::DIRECTORY, '/.*/', ['key' => 'name']);
    $abandoned_files = array_diff_key($files, $files_to_keep);

    foreach ($abandoned_files as $uri => $data) {
      $this->fileSystem->delete($data->uri);
    }
  }

  public function removeCssFiles() {
    // @todo which fits best?
    // $this->fileSystem->rmdir(self::DIRECTORY);
    $this->fileSystem->deleteRecursive(self::DIRECTORY);

    \Drupal::service('library.discovery')->clearCachedDefinitions();
  }

  public function getLibrary(GenericLayout $layout): array {
    $css_uri = self::getCssUri($layout);

    if (!file_exists($css_uri)) {
      $this->storeCssFile($layout);
    }

    return [
      'css' => [
        'layout' => [
          $css_uri => [
            'preprocess' => false,
          ]
        ]
      ]
    ];
  }

  public function getLibraries(): array {
    $libraries = [];

    foreach ($this->getGenericLayoutSections() as $section) {
      $layout = $section->getLayout();
      $libraries[$layout->getInstance()] = self::getLibrary($layout);
    }

    // Now is a cheap and reliable moment to get rid of abandoned css files
    $this->removeAbandonedCssFiles($libraries);

    return $libraries;
  }

  /**
   * If the LayoutSettings scheme changes, this helper makes it easy, to update existing Layouts accordingly.
   * @todo due to the need of changing all Settings in normal operation (f.e. when changing the breakpoint group),
   * the function may be moved to GridLayoutManager
   * This is typically followed by a call of clearLayoutTempstoreRepositories() to also get rid of invalid configurations
   * in currenlty edited layouts
   *
   * Example updpate converting each GenericLayout's key 'uuid' to 'instance':
   *
   * \Drupal::service('generic_layout.manager')->updateLayoutSettings( function(&$settings) {
   *   if(!isset($settings['instance'])) {
   *     $settings['instance'] = $settings['uuid'];
   *   }
   *   unset($settings['uuid']);
   * });
   *
   * @param $callback
   * @return void
   */
  public function updateLayoutSettings(callable $callable, bool $clear_cache = TRUE): void {
    $sections = [];
    $map = $this->getLayoutBuilderEntityMap();
    foreach ($map as $entity_type_id => $ids) {
      $entityStorage = $this->entityTypeManager->getStorage($entity_type_id);

      foreach ($entityStorage->loadMultiple($ids) as $entity) {
        if ($section_storage = $this->getSectionStorageForEntity($entity)) {
          $sections = $section_storage->getSections();
          $gl_sections = GenericLayoutHelper::extractGenericLayoutSections($sections);

          // Invoke user function
          // @todo use 'Drupal Batch API' cause a site may have dozens of Layouts
          // @see https://www.drupal.org/docs/drupal-apis/batch-api/batch-api-overview
          foreach ($gl_sections as &$section) {
            call_user_func($callable, $section);
          }
          $entity->save();
        }
      }
    }

    if ($clear_cache) {
      $this->libraryDiscoveryCollector->delete('generic_layout');
    }
  }

  /**
   * Deletes all layout builder layouts in transitional edit state
   * Usefull to get rid of depricated/invalid/broken config keys
   */
  public function clearLayoutTempstoreRepositories(): int {
    // delete in database because an API solution does not exist.
    // See below code snippet
    $connection = \Drupal::service('database');
    $query = $connection->delete('key_value_expire')->condition('collection', 'tempstore.shared.layout_builder.section_storage.%', "LIKE");
    $count = $query->execute();

    return $count;

    // @see \Drupal\layout_builder\SectionStorageInterface->getStorageType()
    foreach(['defaults', 'overrides'] as $storage_type) {
      $collection = 'layout_builder.section_storage.' . $storage_type;
      $tempstore = \Drupal::service('tempstore.shared')->get($collection);

      // deleteAll API for tempstore does not exist
      // @see https://www.drupal.org/project/drupal/issues/2475719
      // $tempstore->deleteAll();
    }
  }
}
